// Pairing  (STL )
// (c) Proteus     lawnmower-man@mail.ru

#include<stdio.h>
#include<list>

//   
#define MODE_TWOPASS	1	//!<   - 
#define MODE_MULTIPASS	2	//!<   - 
#define POP_MODE MODE_MULTIPASS //MODE_TWOPASS


//!  
struct node_st{
	node_st(int);								
	std::list<node_st*>list;					//!<   
	int value;									//!<  
	int GetH();									//!<   
	static bool Compare(node_st*,node_st*);		//!<   
	static node_st *Merge(node_st*,node_st*);	//!<   
	void Dump(int depth);						//!<  
	void Clear();								//!< 
	node_st *PopFirst();						//!<   
	typedef std::list<node_st*>::iterator it;	//!< 
};


//!  
class PairHeap{
	node_st* root;					//!<   
	void Insert(node_st*);			//!<     
public:
	PairHeap();
	~PairHeap();
	void Insert(int v);				//!<    
	bool isEmpty();					//!<   
	int  Pop();						//!<   
	void Dump();					//!<  
};

//! \param v  
node_st::node_st(int v)
{
	value=v;
}


//!   ,      
node_st *node_st::Merge(node_st *node1,node_st *node2)
{
	node_st *node;
	if(!node1) return node2;
	if(!node2) return node1;
	if(node1->value<node2->value) node=node1,node->list.push_front(node2);
	else node=node2,node->list.push_front(node1);
	return node;
}


bool node_st::Compare(node_st *node1,node_st *node2)
{
	return node1->GetH()==node2->GetH();
}

int node_st::GetH()
{
	return list.size();
}

void node_st::Clear()
{
	for(node_st::it it=list.begin(); it!=list.end(); it++) (*it)->Clear();
	delete this;
}

//! \param depth  
void node_st::Dump(int depth)
{
	for(int n=0; n<depth; n++) putchar(' ');
	printf("[%d]\n",value);
	for(node_st::it it=list.begin(); it!=list.end(); it++) (*it)->Dump(depth+2);
}

//! \result   (0 -   )
node_st* node_st::PopFirst()
{
	if(list.empty()) return 0;
	node_st *node=list.front();
	list.pop_front();
	return node;
}



///==========================================================================


PairHeap::PairHeap()
{
	root=0;
}


PairHeap::~PairHeap()
{
	if(root) root->Clear();
}

//! \param v  
void PairHeap::Insert(int v)
{
	Insert(new node_st(v));
}

void PairHeap::Insert(node_st *node)
{
	if(!root) root=node;
	else root=node_st::Merge(root,node);
}


int PairHeap::Pop()
{
	int v=root->value;
	node_st *temp=root;
#if POP_MODE==MODE_TWOPASS
	node_st::it it;
	std::list<node_st*>list1;
	while(!root->list.empty()) list1.push_back(node_st::Merge(root->PopFirst(),root->PopFirst()));
	root=0;
	while(!list1.empty()) root=node_st::Merge(root,list1.back()),list1.pop_back();
#elif POP_MODE==MODE_MULTIPASS
	while(root->list.size()>1) root->list.push_back(node_st::Merge(root->PopFirst(),root->PopFirst()));
	if(root->list.empty()) root=0; else root=root->list.front();
#elif
#  error "mode"
#endif
	delete temp;
	return v;

}

bool PairHeap::isEmpty()
{
	return !root;
}


void PairHeap::Dump()
{
	printf("[dump]\n");
	if(root) root->Dump(0);
}


#define SIZE 30

void main()
{
	PairHeap heap;
	for(int n=0; n<SIZE; n++) heap.Insert(rand()%30);
	heap.Dump();
	heap.Pop();
	heap.Dump();
	while(!heap.isEmpty()) printf("(%d)",heap.Pop()); putchar('\n');
	printf("[press enter]\n");
	getchar();
}
